@@ -579,31 +579,32 @@ fork_and_print_environment (void)
 		exit (0);
 	}
 	
-	if (!run_daemonized) 
-		return;
+	if (run_daemonized) { 
+		
+		/* Double fork if need to daemonize properly */
+		pid = fork ();
 	
-	/* Double fork if need to daemonize properly */
-	pid = fork ();
+		if (pid != 0) {
+
+			/* Here we are in the intermediate child process */
+				
+			/* 
+			 * This process exits, so that the final child will inherit 
+			 * init as parent to avoid zombies
+			 */
+			if (pid == -1)
+				exit (1);
 	
-	if (pid != 0) {
-
-		/* Here we are in the intermediate child process */
-			
-		/* 
-		 * This process exits, so that the final child will inherit 
-		 * init as parent to avoid zombies
-		 */
-		if (pid == -1)
-			exit (1);
-
-		/* We've done two forks. Now we know the PID */
-		print_environment (pid);
-			
-		/* The intermediate child exits */
-		exit (0);
+			/* We've done two forks. Now we know the PID */
+			print_environment (pid);
+				
+			/* The intermediate child exits */
+			exit (0);
+		}
+		
 	}
 
-	/* Here we are in the resulting daemon process. */
+	/* Here we are in the resulting daemon or background process. */
 
 	for (i = 0; i < 3; ++i) {
 		fd = open ("/dev/null", O_RDONLY);
@@ -612,19 +613,9 @@ fork_and_print_environment (void)
 	}
 }
 
-gboolean
-gkr_daemon_complete_initialization(void)
+static gboolean
+gkr_daemon_complete_initialization_steps (void)
 {
-	/*
-	 * Sometimes we don't initialize the full daemon right on 
-	 * startup. When run with --login is one such case.
-	 */
-	
-	if (initialization_completed) {
-		g_message ("The daemon was already initialized.");
-		return TRUE;
-	}
-	
 	/* Initialize new style PKCS#11 components */
 	if (!gkr_pkcs11_daemon_initialize ())
 		return FALSE;
@@ -649,6 +640,28 @@ gkr_daemon_complete_initialization(void)
 	return TRUE;
 }
 
+gboolean
+gkr_daemon_complete_initialization (void)
+{
+	/*
+	 * Sometimes we don't initialize the full daemon right on 
+	 * startup. When run with --login is one such case.
+	 */
+	
+	if (initialization_completed) {
+		g_message ("The daemon was already initialized.");
+		return TRUE;
+	}
+
+	/* Set this early so that two initializations don't overlap */
+	initialization_completed = TRUE;
+
+	/* But then set it back if initializing fails */
+	initialization_completed = gkr_daemon_complete_initialization_steps ();
+
+	return initialization_completed;
+}
+
 int
 main (int argc, char *argv[])
 {
