@@ -30,8 +30,74 @@
 #include "gnome-keyring-private.h"
 #include "gnome-keyring-memory.h"
 
+#include "egg/egg-secure-memory.h"
+
 /* Functions used by both the library and the daemon */
 
+/* 
+ * A list of all the environment variables the daemon can
+ * possibly send out when it starts. 
+ */
+const gchar *GNOME_KEYRING_OUT_ENVIRONMENT[] = {
+	"SSH_AUTH_SOCK",
+	"GNOME_KEYRING_SOCKET",
+	"GNOME_KEYRING_PID",
+	"SSH_AGENT_PID",
+	NULL
+};
+
+/*
+ * A list of all the environment variables the daemon 
+ * is interested in from clients if it was started 
+ * early before these environment variables were set.
+ */
+const gchar *GNOME_KEYRING_IN_ENVIRONMENT[] = {
+	"DISPLAY",
+	"DBUS_SESSION_BUS_ADDRESS",
+	"DESKTOP_AUTOSTART_ID",
+	"ICEAUTHORITY",
+	"LANG",
+	"XAUTHORITY",
+	"XAUTHLOCALHOSTNAME",
+	"XDG_SESSION_COOKIE",
+	NULL
+};
+
+gchar** 
+gnome_keyring_build_environment (const gchar **names)
+{
+	GArray *array = g_array_sized_new (TRUE, TRUE, sizeof (gchar*), 8);
+	const gchar *value;
+	const gchar **name;
+	gchar *env;
+	
+	/* Transform them into NAME=VALUE pairs */
+	for (name = names; *name; ++name) {
+		value = g_getenv (*name);
+		if (value) {
+			env = g_strdup_printf ("%s=%s", *name, value);
+			g_array_append_val (array, env);
+		}
+	}
+
+	return (gchar**)g_array_free (array, FALSE);
+}
+
+void 
+gnome_keyring_apply_environment (gchar **envp)
+{
+	gchar **e, **parts;
+	
+	g_return_if_fail (envp);
+	
+	for (e = envp; *e; ++e) {
+		parts = g_strsplit (*e, "=", 2);
+		if (parts && parts[0] && parts[1])
+			g_setenv (parts[0], parts[1], TRUE);
+		g_strfreev (parts);
+	}
+}
+
 /**
  * gnome_keyring_free_password:
  * @password: the password to be freed
@@ -43,29 +109,7 @@
 void
 gnome_keyring_free_password (gchar *password)
 {
-	volatile char *vp;
-	size_t len;
-	
-	if (!password)
-		return;
-		
-	/*
-	 * If we're using unpageable 'secure' memory, then the free call
-	 * should zero out the memory, but because on certain platforms 
-	 * we may be using normal memory, zero it out here just in case.
-	 */
-	 
-	/* Defeats some optimizations */		
-	len = strlen (password);
-	memset (password, 0xAA, len);
-	memset (password, 0xBB, len);
-
-	/* Defeats others */
-        vp = (volatile char*)password;
-        while (*vp) 
-        	*(vp++) = 0xAA;
-	
-	gnome_keyring_memory_free (password);
+	egg_secure_strfree (password);
 }
 
 /**
@@ -111,14 +155,16 @@ gnome_keyring_result_to_message (GnomeKeyringResult res)
 		return _("Access Denied");
 	case GNOME_KEYRING_RESULT_NO_KEYRING_DAEMON:
 		return _("The gnome-keyring-daemon application is not running.");
-	case GNOME_KEYRING_RESULT_NO_SUCH_KEYRING:
-		return _("The keyring has already been unlocked.");
 	case GNOME_KEYRING_RESULT_IO_ERROR:
 		return _("Error communicating with gnome-keyring-daemon");
 	case GNOME_KEYRING_RESULT_ALREADY_EXISTS:
 		return _("A keyring with that name already exists");	
 	case GNOME_KEYRING_RESULT_BAD_ARGUMENTS:
 		return _("Programmer error: The application sent invalid data.");
+	case GNOME_KEYRING_RESULT_NO_MATCH:
+		return _("No matching results"); 
+	case GNOME_KEYRING_RESULT_NO_SUCH_KEYRING:
+		return _("A keyring with that name does not exist.");
 	
 	/* 
 	 * This would be a dumb message to display to the user, we never return 
@@ -356,7 +402,7 @@ gnome_keyring_item_info_copy (GnomeKeyringItemInfo *item_info)
 	memcpy (copy, item_info, sizeof (GnomeKeyringItemInfo));
 
 	copy->display_name = g_strdup (copy->display_name);
-	copy->secret = gnome_keyring_memory_strdup (copy->secret);
+	copy->secret = egg_secure_strdup (copy->secret);
 	
 	return copy;
 }
