@@ -35,6 +35,8 @@
 
 #include "egg/egg-hex.h"
 
+#include "pkcs11/pkcs11i.h"
+
 #include <glib/gstdio.h>
 
 #include <libtasn1.h>
@@ -182,7 +184,7 @@ identifier_for_object (GckObject *object)
 	}
 	
 	/* Build up the identifier */
-	identifier = g_strconcat (name ? "-" : "", name, ext, NULL);
+	identifier = g_strconcat (name, ext, NULL);
 	g_strdelimit (identifier, UNWANTED_IDENTIFIER_CHARS, '_');
 
 	g_free (name);
@@ -336,14 +338,14 @@ begin_write_state (GckUserStorage *self, GckTransaction *transaction)
 		return TRUE;
 	}
 	
-	gck_transaction_add (transaction, self, complete_write_state, NULL);
-	self->transaction = g_object_ref (transaction);
-	
 	/* Lock file for the transaction */
 	self->read_fd = begin_lock_file (self, transaction);
 	if (self->read_fd == -1)
 		return FALSE;
 
+	gck_transaction_add (transaction, self, complete_write_state, NULL);
+	self->transaction = g_object_ref (transaction);
+
 	/* Open the new file */
 	g_assert (self->write_fd == -1);
 	self->write_path = g_strdup_printf ("%s.XXXXXX", self->filename);
@@ -358,7 +360,33 @@ begin_write_state (GckUserStorage *self, GckTransaction *transaction)
 }
 
 static gboolean
-begin_write_state_and_refresh (GckUserStorage *self, GckTransaction *transaction)
+complete_modification_state (GckTransaction *transaction, GObject *object, gpointer unused)
+{
+	GckUserStorage *self = GCK_USER_STORAGE (object);
+	GckDataResult res;
+	
+	if (!gck_transaction_get_failed (transaction)) {
+		res = gck_data_file_write_fd (self->file, self->write_fd, self->login);
+		switch(res) {
+		case GCK_DATA_FAILURE:
+		case GCK_DATA_UNRECOGNIZED:
+			g_warning ("couldn't write to temporary store file: %s", self->write_path);
+			return FALSE;
+		case GCK_DATA_LOCKED:
+			g_warning ("couldn't encrypt temporary store file: %s", self->write_path);
+			return FALSE;
+		case GCK_DATA_SUCCESS:
+			break;
+		default:
+			g_assert_not_reached ();
+		}
+	}
+	
+	return TRUE;
+}
+
+static gboolean
+begin_modification_state (GckUserStorage *self, GckTransaction *transaction)
 {
 	GckDataResult res;
 	struct stat sb;
@@ -397,6 +425,9 @@ begin_write_state_and_refresh (GckUserStorage *self, GckTransaction *transaction
 		}
 	}
 	
+	/* Write out the data once completed with modifications */
+	gck_transaction_add (transaction, self, complete_modification_state, NULL);
+	
 	return TRUE;
 }
 
@@ -421,12 +452,61 @@ take_object_ownership (GckUserStorage *self, const gchar *identifier, GckObject
 	gck_manager_register_object (self->manager, object);
 }
 
+static gboolean
+check_object_hash (GckUserStorage *self, const gchar *identifier, const guchar *data, gsize n_data)
+{
+	gconstpointer value;
+	GckDataResult res;
+	gboolean result;
+	gsize n_value;
+	gchar *digest;
+	
+	g_assert (GCK_IS_USER_STORAGE (self));
+	g_assert (identifier);
+	g_assert (data);
+	
+	digest = g_compute_checksum_for_data (G_CHECKSUM_SHA1, data, n_data);
+	g_return_val_if_fail (digest, FALSE);
+	
+	res = gck_data_file_read_value (self->file, identifier, CKA_GNOME_INTERNAL_SHA1, &value, &n_value);
+	g_return_val_if_fail (res == GCK_DATA_SUCCESS, FALSE);
+	
+	result = (strlen (digest) == n_value && memcmp (digest, value, n_value) == 0);
+	g_free (digest);
+	
+	return result;
+}
+
+static void
+store_object_hash (GckUserStorage *self, GckTransaction *transaction, const gchar *identifier, 
+                   const guchar *data, gsize n_data)
+{
+	GckDataResult res;
+	gchar *digest;
+	
+	g_assert (GCK_IS_USER_STORAGE (self));
+	g_assert (GCK_IS_TRANSACTION (transaction));
+	g_assert (identifier);
+	g_assert (data);
+	
+	digest = g_compute_checksum_for_data (G_CHECKSUM_SHA1, data, n_data);
+	if (digest == NULL) {
+		gck_transaction_fail (transaction, CKR_GENERAL_ERROR);
+		g_return_if_reached ();
+	}
+	
+	res = gck_data_file_write_value (self->file, identifier, CKA_GNOME_INTERNAL_SHA1, digest, strlen (digest));
+	g_free (digest);
+	
+	if (res != GCK_DATA_SUCCESS)
+		gck_transaction_fail (transaction, CKR_GENERAL_ERROR);
+}
+
 static void 
 data_file_entry_added (GckDataFile *store, const gchar *identifier, GckUserStorage *self)
 {
 	GError *error = NULL;
 	GckObject *object;
-	GckDataResult res;
 	gboolean ret;
 	guchar *data;
 	gsize n_data;
@@ -460,32 +540,24 @@ data_file_entry_added (GckDataFile *store, const gchar *identifier, GckUserStora
 		return;
 	}
 	
+	/* Make sure that the object wasn't tampered with */
+	if (!check_object_hash (self, identifier, data, n_data)) {
+		g_message ("file in user store doesn't match hash: %s", identifier);
+		return;
+	}
+	
 	/* Create a new object for this identifier */
 	object = g_object_new (type, "unique", identifier, NULL);
 	g_return_if_fail (GCK_IS_SERIALIZABLE (object));
 	g_return_if_fail (GCK_SERIALIZABLE_GET_INTERFACE (object)->extension);
 
 	/* And load the data into it */
-	res = gck_serializable_load (GCK_SERIALIZABLE (object), self->login, data, n_data);
-	g_free (data);
-	
-	switch (res) {
-	case GCK_DATA_FAILURE:
-		g_message ("failed to load file in user store: %s", identifier);
-		return;
-	case GCK_DATA_UNRECOGNIZED:
-		g_message ("invalid or unparsable file in user store: %s", identifier);
-		return;
-	case GCK_DATA_LOCKED:
-		g_message ("file is locked with unknown password: %s", identifier);
-		return;
-	case GCK_DATA_SUCCESS:
+	if (gck_serializable_load (GCK_SERIALIZABLE (object), self->login, data, n_data)) 
 		take_object_ownership (self, identifier, object);
-		break;
-	default:
-		g_assert_not_reached ();
-	}
+	else 
+		g_message ("failed to load file in user store: %s", identifier);
 	
+	g_free (data);
 	g_object_unref (object);
 }
 
@@ -521,17 +593,16 @@ data_file_entry_removed (GckDataFile *store, const gchar *identifier, GckUserSto
 }
 
 static void
-relock_object (GckTransaction *transaction, const gchar *path, 
+relock_object (GckUserStorage *self, GckTransaction *transaction, const gchar *path, 
                const gchar *identifier, GckLogin *old_login, GckLogin *new_login)
 {
 	GError *error = NULL;
 	GckObject *object;
-	GckDataResult res;
 	guchar *data;
 	gsize n_data;
 	GType type;
-	CK_RV rv;
 	
+	g_assert (GCK_IS_USER_STORAGE (self));
 	g_assert (GCK_IS_TRANSACTION (transaction));
 	g_assert (identifier);
 	g_assert (path);
@@ -556,70 +627,53 @@ relock_object (GckTransaction *transaction, const gchar *path,
 	
 	/* Read in the data for the object */
 	if (!g_file_get_contents (path, (gchar**)&data, &n_data, &error)) {
-		g_message ("couldn't relock file in user store: %s: %s", identifier,
+		g_message ("couldn't load file in user store in order to relock: %s: %s", identifier,
 		           error && error->message ? error->message : "");
 		g_clear_error (&error);
 		g_object_unref (object);
+		gck_transaction_fail (transaction, CKR_GENERAL_ERROR);
+		return;
+	}
+	
+	/* Make sure the data matches the hash */
+	if (!check_object_hash (self, identifier, data, n_data)) {
+		g_message ("file in data store doesn't match hash: %s", identifier);
+		gck_transaction_fail (transaction, CKR_GENERAL_ERROR);
 		return;
 	}
 	
 	/* Load it into our temporary object */
-	res = gck_serializable_load (GCK_SERIALIZABLE (object), old_login, data, n_data);
-	g_free (data);
-	
-	switch (res) {
-	case GCK_DATA_FAILURE:
-	case GCK_DATA_UNRECOGNIZED:
+	if (!gck_serializable_load (GCK_SERIALIZABLE (object), old_login, data, n_data)) {
 		g_message ("unrecognized or invalid user store file: %s", identifier);
-		rv = CKR_FUNCTION_FAILED;
-		break;
-	case GCK_DATA_LOCKED:
-		g_message ("old login is invalid for user store file: %s", identifier);
-		rv = CKR_PIN_INCORRECT;
-		break;
-	case GCK_DATA_SUCCESS:
-		rv = CKR_OK;
-		break;
-	default:
-		g_assert_not_reached ();
-	}
-	
-	if (rv != CKR_OK) {
-		gck_transaction_fail (transaction, rv);
+		gck_transaction_fail (transaction, CKR_FUNCTION_FAILED);
+		g_free (data);
 		g_object_unref (object);
 		return;
-	}
+	} 
+	
+	g_free (data);
+	data = NULL;
 		
 	/* Read it out of our temporary object */
-	res = gck_serializable_save (GCK_SERIALIZABLE (object), new_login, &data, &n_data);
-	g_object_unref (object);
-	
-	switch (res) {
-	case GCK_DATA_FAILURE:
-	case GCK_DATA_UNRECOGNIZED:
+	if (!gck_serializable_save (GCK_SERIALIZABLE (object), new_login, &data, &n_data)) {
 		g_warning ("unable to serialize data with new login: %s", identifier);
-		rv = CKR_GENERAL_ERROR;
-		break;
-	case GCK_DATA_LOCKED:
-		g_message ("new login is invalid for user store file: %s", identifier);
-		rv = CKR_PIN_INVALID;
-		break;
-	case GCK_DATA_SUCCESS:
-		rv = CKR_OK;
-		break;
-	default:
-		g_assert_not_reached ();
-	}
-
-	if (rv != CKR_OK) {
-		gck_transaction_fail (transaction, rv);
+		gck_transaction_fail (transaction, CKR_GENERAL_ERROR);
+		g_object_unref (object);
 		g_free (data);
 		return;
 	}
 	
+	g_object_unref (object);
+	
 	/* And write it back out to the file */
 	gck_transaction_write_file (transaction, path, data, n_data);
+	
+	/* Create and save the hash here */
+	if (!gck_transaction_get_failed (transaction))
+		store_object_hash (self, transaction, identifier, data, n_data);
+	
 	g_free (data);
+
 }
 
 typedef struct _RelockArgs {
@@ -648,7 +702,7 @@ relock_each_object (GckDataFile *file, const gchar *identifier, gpointer data)
 		return;
 
 	path = g_build_filename (args->self->directory, identifier, NULL);
-	relock_object (args->transaction, path, identifier, args->old_login, args->new_login);
+	relock_object (args->self, args->transaction, path, identifier, args->old_login, args->new_login);
 	g_free (path);
 }
 
@@ -705,18 +759,6 @@ refresh_with_login (GckUserStorage *self, GckLogin *login)
 	return rv;
 }
 
-static void
-set_storage_login (GckUserStorage *self, GckLogin *login)
-{
-	g_assert (GCK_IS_USER_STORAGE (self));
-	if (login != NULL)
-		g_object_ref (login);
-	if (self->login)
-		g_object_unref (self->login);
-	self->login = login;
-	g_object_notify (G_OBJECT (self), "login");
-}
-
 /* -----------------------------------------------------------------------------
  * OBJECT 
  */
@@ -838,6 +880,9 @@ gck_user_storage_init (GckUserStorage *self)
 	/* Each one owns the key and contains weak ref to other's key as its value */
 	self->object_to_identifier = g_hash_table_new_full (g_direct_hash, g_direct_equal, dispose_unref_object, NULL);
 	self->identifier_to_object = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+	
+	self->read_fd = -1;
+	self->write_fd = -1;
 }
 
 static void
@@ -1015,7 +1060,7 @@ gck_user_storage_create (GckUserStorage *self, GckTransaction *transaction, GckO
 	}
 
 	/* Hook ourselves into the transaction */
-	if (!begin_write_state_and_refresh (self, transaction))
+	if (!begin_modification_state (self, transaction))
 		return;
 
 	/* Create an identifier guaranteed unique by this transaction */
@@ -1025,8 +1070,16 @@ gck_user_storage_create (GckUserStorage *self, GckTransaction *transaction, GckO
 		g_return_if_reached ();
 	}
 	
+	/* We don't want to get signals about this item being added */
+	g_signal_handlers_block_by_func (self->file, data_file_entry_added, self);
+	g_signal_handlers_block_by_func (self->file, data_file_entry_changed, self);
+	
 	res = gck_data_file_create_entry (self->file, identifier, 
 	                                  is_private ? GCK_DATA_FILE_SECTION_PRIVATE : GCK_DATA_FILE_SECTION_PUBLIC);
+	
+	g_signal_handlers_unblock_by_func (self->file, data_file_entry_added, self);
+	g_signal_handlers_unblock_by_func (self->file, data_file_entry_changed, self);
+
 	switch(res) {
 	case GCK_DATA_FAILURE:
 	case GCK_DATA_UNRECOGNIZED:
@@ -1044,14 +1097,17 @@ gck_user_storage_create (GckUserStorage *self, GckTransaction *transaction, GckO
 	}
 	
 	/* Serialize the object in question */
-	res = gck_serializable_save (GCK_SERIALIZABLE (object), is_private ? self->login : NULL, &data, &n_data);
-	if (res != GCK_DATA_SUCCESS) {
+	if (!gck_serializable_save (GCK_SERIALIZABLE (object), is_private ? self->login : NULL, &data, &n_data)) {
 		gck_transaction_fail (transaction, CKR_FUNCTION_FAILED);
 		g_return_if_reached ();
 	}
 
 	path = g_build_filename (self->directory, identifier, NULL);
 	gck_transaction_write_file (transaction, path, data, n_data);
+	
+	/* Make sure we write in the object hash */
+	if (!gck_transaction_get_failed (transaction))
+		store_object_hash (self, transaction, identifier, data, n_data);
 
 	/* Now we decide to own the object */
 	if (!gck_transaction_get_failed (transaction))
@@ -1078,7 +1134,7 @@ gck_user_storage_destroy (GckUserStorage *self, GckTransaction *transaction, Gck
 	identifier = g_hash_table_lookup (self->object_to_identifier, object);
 	g_return_if_fail (identifier);
 	
-	if (!begin_write_state_and_refresh (self, transaction))
+	if (!begin_modification_state (self, transaction))
 		return;
 
 	/* First actually delete the file */
@@ -1164,8 +1220,13 @@ gck_user_storage_relock (GckUserStorage *self, GckTransaction *transaction,
 	args.new_login = new_login;
 	gck_data_file_foreach_entry (file, relock_each_object, &args);
 	
-	if (!gck_transaction_get_failed (transaction))
-		set_storage_login (self, new_login);
+	if (!gck_transaction_get_failed (transaction) && self->login) {
+		if (new_login)
+			g_object_ref (new_login);
+		g_object_unref (self->login);
+		self->login = new_login;
+		g_object_notify (G_OBJECT (self), "login");
+	}
 	
 	g_object_unref (file);
 }
@@ -1181,11 +1242,23 @@ gck_user_storage_unlock (GckUserStorage *self, GckLogin *login)
 	if (self->login)
 		return CKR_USER_ALREADY_LOGGED_IN;
 	
+	self->login = login;
+	
 	rv = refresh_with_login (self, login);
 	if (rv == CKR_USER_NOT_LOGGED_IN)
-		return CKR_PIN_INCORRECT;
-	else if (rv == CKR_OK)
-		set_storage_login (self, login);
+		rv = CKR_PIN_INCORRECT;
+	
+	/* Take on new login for good */
+	if (rv == CKR_OK) {
+		g_assert (self->login == login);
+		if (self->login)
+			g_object_ref (self->login);
+		g_object_notify (G_OBJECT (self), "login");
+		
+	/* Failed, so keep our previous NULL login */
+	} else {
+		self->login = NULL;
+	}
 	
 	return rv;
 }
@@ -1193,6 +1266,7 @@ gck_user_storage_unlock (GckUserStorage *self, GckLogin *login)
 CK_RV
 gck_user_storage_lock (GckUserStorage *self)
 {
+	GckLogin *prev;
 	CK_RV rv;
 	
 	g_return_val_if_fail (GCK_IS_USER_STORAGE (self), CKR_GENERAL_ERROR);
@@ -1201,9 +1275,22 @@ gck_user_storage_lock (GckUserStorage *self)
 	if (!self->login)
 		return CKR_USER_NOT_LOGGED_IN;
 	
+	/* While loading set new NULL login */
+	prev = self->login;
+	self->login = NULL;
+	
 	rv = refresh_with_login (self, NULL);
-	if (rv == CKR_OK)
-		set_storage_login (self, NULL);
+	
+	/* Take on new login for good */
+	if (rv == CKR_OK) {
+		g_object_unref (prev);
+		g_assert (self->login == NULL);
+		g_object_notify (G_OBJECT (self), "login");
+		
+	/* Failed so revert to previous login */
+	} else {
+		self->login = prev;
+	}
 	
 	return rv;
 }
@@ -1228,3 +1315,28 @@ gck_user_storage_get_login (GckUserStorage *self)
 	g_return_val_if_fail (GCK_IS_USER_STORAGE (self), NULL);
 	return self->login;
 }
+
+gulong
+gck_user_storage_token_flags (GckUserStorage *self)
+{
+	gulong flags = 0;
+	CK_RV rv;
+	
+	/* We don't changing SO logins, so always initialized */
+	flags |= CKF_TOKEN_INITIALIZED | CKF_LOGIN_REQUIRED;
+	
+	/* No file has been loaded yet? */
+	if (self->last_mtime == 0) {
+		rv = gck_user_storage_refresh (self);
+		if (rv == CKR_USER_PIN_NOT_INITIALIZED)
+			flags |= CKF_USER_PIN_TO_BE_CHANGED;
+		else if (rv != CKR_OK)
+			g_return_val_if_reached (flags);
+	}
+	
+	/* No private stuff in the file? */
+	if (gck_data_file_have_section (self->file, GCK_DATA_FILE_SECTION_PRIVATE))
+		flags |= CKF_USER_PIN_INITIALIZED;
+	
+	return flags;
+}
